# Copyright (C) 2013 by Aivars Kalvans <aivars.kalvans@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import re
import random
import base64
import logging
import pkgutil

log = logging.getLogger('scrapy.proxies')


class Mode:
    RANDOMIZE_PROXY_EVERY_REQUESTS, RANDOMIZE_PROXY_ONCE, SET_CUSTOM_PROXY, RANDOMIZE_PART_PROXY = range(4)


class RandomProxy(object):
    def __init__(self, settings):
        self.mode = int(settings.get('PROXY_MODE'))
        self.proxy_list = settings.get('PROXY_LIST')
        self.proxy_pkg = settings.get('PROXY_PKG')
        self.proxy_random_size = 1
        if settings.get('PROXY_RANDOM_SIZE'):
            self.proxy_random_size = int(settings.get('PROXY_RANDOM_SIZE'))
        self.chosen_proxies = []
        self.chosen_proxy = ''

        if self.mode == Mode.RANDOMIZE_PROXY_EVERY_REQUESTS or self.mode == Mode.RANDOMIZE_PROXY_ONCE or self.mode == Mode.RANDOMIZE_PART_PROXY:
            if self.proxy_list is None:
                raise KeyError('PROXY_LIST setting is missing')
            self.proxies = {}
            if self.proxy_pkg is None or self.proxy_pkg == 'None':
                fin = open(self.proxy_list)
                lines = fin.readlines()
            else:
                fin = pkgutil.get_data(self.proxy_pkg, self.proxy_list)
                lines = fin.decode("utf-8").splitlines()
            try:
                for line in lines:
                    parts = re.match('(\w+://)([^:]+?:[^@]+?@)?(.+)', line.strip())
                    if not parts:
                        continue

                    # Cut trailing @
                    if parts.group(2):
                        user_pass = parts.group(2)[:-1]
                    else:
                        user_pass = ''

                    self.proxies[parts.group(1) + parts.group(3)] = user_pass
                    log.info('%s %d left' % ('total proxies', len(self.proxies)))
            finally:
                if self.proxy_pkg is None or self.proxy_pkg == 'None':
                    fin.close()
            if self.mode == Mode.RANDOMIZE_PROXY_ONCE:
                self.chosen_proxy = random.choice(list(self.proxies.keys()))
                log.info('%s chosen in %d proxies' % (self.chosen_proxy, len(self.proxies)))
            if self.mode == Mode.RANDOMIZE_PART_PROXY:
                self.chosen_proxies = random.choices(list(self.proxies.keys()), k=self.proxy_random_size)
                log.info('%s chosen in %d proxies' % (self.chosen_proxies, len(self.proxies)))
        elif self.mode == Mode.SET_CUSTOM_PROXY:
            custom_proxy = settings.get('CUSTOM_PROXY')
            self.proxies = {}
            parts = re.match('(\w+://)([^:]+?:[^@]+?@)?(.+)', custom_proxy.strip())
            if not parts:
                raise ValueError('CUSTOM_PROXY is not well formatted')

            if parts.group(2):
                user_pass = parts.group(2)[:-1]
            else:
                user_pass = ''

            self.proxies[parts.group(1) + parts.group(3)] = user_pass
            self.chosen_proxy = parts.group(1) + parts.group(3)
            log.info('%s chosen %d proxies' % (self.chosen_proxy, len(self.proxies)))

    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler.settings)

    def process_request(self, request, spider):
        # allow spider instance to control if enable proxy or not
        if not getattr(spider, 'enable_proxy', False):
            return

        # Don't overwrite with a random one (server-side state for IP)
        if 'proxy' in request.meta:
            if request.meta["exception"] is False:
                # fill authentication
                proxy_address = request.meta['proxy']
                proxy_user_pass = self.proxies[proxy_address]
                basic_auth = 'Basic ' + base64.b64encode(proxy_user_pass.encode()).decode()
                request.headers['Proxy-Authorization'] = basic_auth
                return

        request.meta["exception"] = False
        if len(self.proxies) == 0:
            raise ValueError('All proxies are unusable, cannot proceed')

        if self.mode == Mode.RANDOMIZE_PROXY_EVERY_REQUESTS:
            proxy_address = random.choice(list(self.proxies.keys()))
        elif self.mode == Mode.RANDOMIZE_PART_PROXY:
            proxy_address = random.choice(list(self.chosen_proxies))
        else:
            if getattr(spider, 'proxy_replace', False):
                proxy_address = getattr(spider, 'proxy_replace', False)
            else:
                proxy_address = self.chosen_proxy

        if proxy_address in self.proxies:
            proxy_user_pass = self.proxies[proxy_address]

            if proxy_user_pass:
                request.meta['proxy'] = proxy_address
                basic_auth = 'Basic ' + base64.b64encode(proxy_user_pass.encode()).decode()
                request.headers['Proxy-Authorization'] = basic_auth

                if getattr(spider, 'enable_proxy_slot', False):
                    request.meta['download_slot'] = proxy_address
            else:
                log.debug('Proxy user pass not found')

        log.debug('Using proxy <%s>, %d proxies left' % (
                proxy_address, len(self.proxies)))

    def process_exception(self, request, exception, spider):
        # allow spider instance to control if enable proxy or not
        if not getattr(spider, 'enable_proxy', False):
            return

        if 'proxy' not in request.meta:
            return
        if self.mode == Mode.RANDOMIZE_PROXY_EVERY_REQUESTS or self.mode == Mode.RANDOMIZE_PART_PROXY or self.mode == Mode.RANDOMIZE_PROXY_ONCE:
            proxy = request.meta['proxy']
            # try:
            #     del self.proxies[proxy]
            # except KeyError:
            #     pass
            request.meta["exception"] = True
            if self.mode == Mode.RANDOMIZE_PROXY_ONCE:
                self.chosen_proxy = random.choice(list(self.proxies.keys()))
            log.info('skip failed proxy <%s>, %d proxies left' % (
                proxy, len(self.proxies)))
